#!/usr/bin/perl -w
#
#   Randomly configure the username & team a Folding@Home client should use.
#   Copyright (C) 2007  Sebastiaan Couwenberg <sebastic@xs4all.nl>
#   
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License along
#   with this program; if not, write to the Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Manage a Folding@Home instance to run each new WU for a randomly picked 
# username & team combination in the configuration file.
#
# It uses Dick Howells qd (Queue Dumper) utility to monitor when the F@H
# client, started with -oneunit, has successfully completed a WU.
#
# The F@H clients client.cfg file will be modified to use a new username & team
# combination. And the F@H client will be restarted with the new configuration
# and the -oneunit argument to process only a single WU and quit.
#

use strict;
use Cwd;
use File::Basename;
use Date::Calc;
use XML::Simple;
use Getopt::Mixed "nextOption";

$|=1; # show what's printed without waiting for the newline

my $directory = dirname($0) .'/';
my $fahclient = 'FAH504-Linux.exe';
my @arguments = qw/ -verbosity 9 -oneunit /;

my $logname   = 'random-folder.log';
my $logfile   = $directory.$logname;

my %cfg = ( self      => 'random-folder.pl',
            version   => '0.1.1',
	   
	    directory => $directory,
	    logfile   => $logfile,
	    fahclient => $fahclient,
	    arguments => \@arguments,
	    cli_args  => '',
	    
	    verbose   => 0
          );

# Get commandline parameters
Getopt::Mixed::init("directory=s
                     client=s
                     arguments=s
		     verbose
		     help
		     d>directory
		     c>client
		     a>arguments
		     v>verbose
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	elsif($option eq 'arguments') {
		#$cfg{arguments} = parse_arguments($value);	
		$cfg{cli_args} = $value;	
	}

	$cfg{directory}  = $value if($option eq 'directory');
	$cfg{fahclient}  = $value if($option eq 'client');

	$cfg{verbose}    = 1      if($option eq 'verbose');
}
Getopt::Mixed::cleanup();

$cfg{directory} .= '/' if(substr($cfg{directory}, -1, 1) ne '/');
$cfg{logfile}    = $cfg{directory}.$logname;


# Open the logfile
if(-e $cfg{logfile} && -r $cfg{logfile} && -w $cfg{logfile}) {
	open(LOG, ">", $cfg{logfile}) || die "Error: Cannot open logfile: $cfg{logfile} ($!)";
	close LOG;
	
	if($cfg{verbose}) {
		print_date("$cfg{self} starting.\n");
		print_date("\n");
	}
}
elsif(!-e $cfg{logfile} && -w $cfg{directory}) {
	open(LOG, ">", $cfg{logfile}) || die "Error: Cannot open logfile: $cfg{logfile} ($!)";
	close LOG;
		
	if($cfg{verbose}) {
		print_date("$cfg{self} starting.\n");
		print_date("\n");
	}
}
else {
	print_date("Error: Cannot write logfile: $cfg{logfile}\n");
	print_date("       Check the permissions!\n");

	exit 1;
}


# Change working directory to FAH client directory 
if(-d $cfg{'directory'} && -x $cfg{'directory'}) {
        print_date("Changing working directory to: $cfg{'directory'}\n") if($cfg{verbose});
        Cwd::chdir $cfg{'directory'};
        print_date("\n") if($cfg{verbose});
}
else {
        print_date("Error: Cannot change to directory: $cfg{'directory'}\n");
        exit 1;
}


# Parse the commandline arguments for the FAH client
$cfg{arguments} = parse_arguments($cfg{cli_args});


# Start the F@H client
if(-x $cfg{'fahclient'}) {
	# start the client
	my $pid = start_client();
	
	# wait for the client to terminate, reconfigure the client, and restart it
	manage_client($pid);
}
else {
	print_date("Error: Cannot execute FAH client: $cfg{'fahclient'}\n");
	exit(1);
}


exit(0);


#################################[ Subroutines ]################################

sub terminate {
	my $signal = shift;

	if($cfg{verbose}) {
		print_log("\n");
		print_date("Cought $signal, terminating...\n");
	}

	exit 0;
}

sub print_date {
	my $string = shift;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

	my $message = "[$date $time] $string";
	
	print_log($message);
}

sub print_log {
	my $string = shift;
	
	# print to STDOUT
	print $string;
	
	# print to logfile
	open(LOG, ">>", $cfg{logfile}) || die "Error: Cannot open logfile: $cfg{logfile} ($!)";
	print LOG $string;
	close LOG;
}

sub parse_arguments {
	# parse and sanitize the F@H client commandline parameters
	my $value = shift;
	
	# The F@H client supports the following arguments:
	#
	#  -config      Configure user information
	#  -configonly  Configure user information, then exit
	#  -help        Dispaly this usage screen
	#  -queueinfo   Get information on queued work units
	#  -delete x    Delete item #x from work queue
	#  -send x      Send result #x to server then exit. Use x=all to send all results
	#  -verbosity x Sets the output level, from 1 to 9 (max). The default is 3
	#  -pause       Pause after finishing & trying to send current unit
	#  -oneunit     Exit after completing one unit
	#  -forceasm    Force core assembly optimizations to be used if available
	#  -advmethods  Use new advanced scientific cores and/or work units if available
	#  -freeBSD     Make brandelf system call on downloaded cores.
	#  -openBSD     Make elf2olf system call on downloaded cores.
	# 
	# Only these arguments will be accepted:
	#
	#  -forceasm    Force core assembly optimizations to be used if available
	#  -advmethods  Use new advanced scientific cores and/or work units if available
	#  -freeBSD     Make brandelf system call on downloaded cores.
	#  -openBSD     Make elf2olf system call on downloaded cores.
	
	print_date("Parsing FAH client commandline arguments.\n") if($cfg{verbose});
	
	my $valid = 1;
	my @arguments = @{$cfg{arguments}};

	foreach(split /\s+/, $value) {
		if(!/^-(forceasm|advmethods|freeBSD|openBSD)$/) {
			print_date("Error: Invalid argument: $_\n");
			$valid = 0;
		}
		else {
			print_date("Adding argument: $_\n") if($cfg{verbose});
			push @arguments, $_;
		}
	}

	if($valid) {
		print_date("\n") if($cfg{verbose});
		return \@arguments;
	}
	else {
		print_date("Error: Invalid arguments to start FAH client!\n");
		print_date("Supported arguments:\n");
		print_date(" -forceasm\n");
		print_date(" -advmethods\n");
		print_date(" -freeBSD\n");
		print_date(" -openBSD\n");
		
		exit(1);
	}
}

sub reconfigure_client {
	# change the F@H client config to use a randomly picked username/team combination
	
	if($cfg{verbose}) {
		print_log("\n");
		print_date("Reconfiguring FAH client.\n");
	}
	
	my $file   = $cfg{directory}.'config.xml';
	my $config = XMLin( $file,
                            KeyAttr    => 'id',
			    ForceArray => [ 'account' ]
			  ) || die($@);
	
	if($cfg{verbose}) {
		print_date("\n");
		print_date("Configured accounts:\n");
	}
	
	my @account;
	foreach my $id (sort keys %{$config->{accounts}->{account}}) {
		my $username = $config->{accounts}->{account}->{"$id"}->{username};
		my $team     = $config->{accounts}->{account}->{"$id"}->{team};
		
		if($cfg{verbose}) {
			print_date(" ID      : $id\n");
			print_date(" Username: $username\n");
			print_date(" Team    : $team\n");
			print_date("\n");
		}
		
		push @account, { username => $username, team => $team };
	}
	
	# Pick a random account
	my $random = int(rand($#account));
	if($cfg{verbose}) {
		print_date("Random account: $random\n");
		print_date("      Username: $account[$random]->{username}\n");
		print_date("          Team: $account[$random]->{team}\n");
		print_date("\n");
	}

	my $clientcfg = $cfg{directory}.'client.cfg';
	if(!-e $clientcfg) {
		print_date("Error: FAH client not configured yet!\n");
		print_date("First run: $cfg{fahclient} -configonly\n");
		
		exit(1);
	}

	print_date("Modifying client.cfg\n") if($cfg{verbose});
	if(-r $clientcfg && -w $clientcfg) {
		my $data = '';
		open(F, $clientcfg) || die("Error: Cannot open file: $clientcfg ($!)\n");
		while(<F>) {
			#username=Anonymous
			if(/^username=(.*?)\s*$/) {
				my $user = $1;
				
				my $line = $_;
				if($line =~ s/\Q$user\E/$account[$random]->{username}/) {}
				
				$data .= $line;
			}
			#team=0
			elsif(/^team=(.*?)\s*$/) {
				my $team = $1;
				
				my $line = $_;
				if($line =~ s/\Q$team\E/$account[$random]->{team}/) {}
				
				$data .= $line;
			}
			else {
				$data .= $_;
			}
		}
		close F;
			
		open(F, ">", $clientcfg) || die("Error: Cannot write file: $clientcfg ($!)\n");
		print F $data;
		close F;
		
		print_date("\n") if($cfg{verbose});
	}
	else {
		print_date("Error: Cannot read and/or write $clientcfg\n");
		
		exit(1);
	}
}

sub start_client {
	# run the F@H client

	print_date("Forking...\n\n") if($cfg{verbose});
	my $pid = fork();
	if($pid) {
		# parent process
		print_date("Parent PID: $$\n") if($cfg{verbose});
		
		return $pid;
	} 
	elsif($pid == 0) {
		# child process
		print_date("Child PID: $$\n") if($cfg{verbose});	
	
		if($cfg{verbose}) {
			print_date("Starting FAH client: ./$cfg{fahclient}");
			foreach(@{$cfg{arguments}}) {
				print_log(" $_");
			}
			print_log("\n\n");
		}

		# Run the F@H client discarding its output,
		# the clients output is written to FAHlog.txt
		#`./$cfg{fahclient} @{$cfg{arguments}} > /dev/null 2>&1`;
		my $status = system("./$cfg{fahclient} @{$cfg{arguments}} > /dev/null 2>&1");
		
		if($cfg{verbose}) {
			print_log("\n");
			print_date("Child PID: $$\n");
			print_date("FAH client exited with status: $status\n\n");
		}
		
		# quit the child process when the F@H client has quit
		exit 0;
	} 
	else {
		print_date("Error: Cannot fork! ($!)\n");

		exit(1);
	}
}

sub is_running {
	my $process = shift;

	if($process =~ /^\s*$/) {
		print_date("Error: No process specified!\n");
		exit(1);
	}

	# check if the process is running
	print_date("Checking if process is running: $process\n") if($cfg{verbose});
	
	my $cmd    = "ps aux | egrep \"$process\" | egrep -v \"grep|random-folder\" |";
	open(CMD, $cmd) || die("Error: Cannot open pipe: '$cmd' ($!)");
	my @output = <CMD>;
	close CMD;
	
	if($#output >= 0) {
		# one or more instances of the process is running
		print_date("Yes, process is running\n") if($cfg{verbose});

		return 1;
	}
	else {
		print_date("No, process is not running\n") if($cfg{verbose});

		return 0;
	}
}

sub client_has_quit {
	# check if the child process which started the client has quit
	# and if all FAH client processes have quit,
	# when started with -oneunit the client should quit
	# after completing a single work unit
	# 
	# Linux CPU client processes: FAH50*-Linux.exe, FahCore_*.exe
	# Linux SMP client processes: fah5, mpiexec, FahCore_*.exe
	
	my $pid = shift;
	
	my $child_running = kill 0, $pid;
	if($cfg{verbose}) {
		print_date("Checking if process is running: child process $pid\n");
		if($child_running) {
		        print_date("Yes, process is running\n");
		}
		else {
			print_date("No, process is not running\n");
		}
	}

	my @processes = ();
	
	# v5.02: FAH502-Linux.exe
	# v5.04: FAH504-Linux.exe
	if($cfg{fahclient} =~ /^FAH5\d{2}-Linux.exe$/) {
		push @processes, ('FAH5.*-Linux.exe', 'FahCore_.*.exe');
	}
	# v5.91: fah5 
	elsif($cfg{fahclient} =~ /^fah5$/) {
		push @processes, ('fah5', 'mpiexec', 'FahCore_.*.exe');
	}
	else {
		print_date("Error: Unsupported FAH client: $cfg{fahclient}\n");
		
		exit(1);
	}

	# if the client type cannot be determined,
	# assume the client is still running
	if($#processes >= 0) {
		my $running = -1;
		foreach(@processes) {
			if(is_running($_)) {
				$running = 1;
				
				# not all processes have quit
				return 0;
			}
			else {
				$running = 0;
			}
		}

		if($running == 0 && !$child_running) {
			# all processes have quit,
			return 1;
		}
	}

	return 0;
}

sub manage_client {
	# wait for the client to terminate, reconfigure it, 
	# and restart the client with the new configuration
	my $pid = shift;

	print_date("Entering client management loop.\n\n") if($cfg{verbose});
	while(1) {
		# indefinetly wait for the client to finish the WU and terminate
		#
		# adjust the sleep time between checks semi-intelligently:
		#  check every 60 minutes if progress <  80%
		#  check every 30 minutes if progress <  95%
		#  check every  5 minutes if progress >= 95%
		
		print_date("Checking WU status & progress ...\n") if($cfg{verbose});
		my %current_wu = get_current_wu_info();
		
		# sleep 60 minutes by default
		my $sleep = 60 * 60;
		
		# <status>folding now</status>
		print_date("Status  : $current_wu{status}\n") if($cfg{verbose});
		
		# <progress>30%</progress>
		print_date("Progress: $current_wu{progress}\n") if($cfg{verbose} && $current_wu{progress});
		if($current_wu{progress} && $current_wu{progress} =~ /^(\d+(\.\d+)?)\%$/) {
			my $progress = $1;

			if($progress >= 95) {
				$sleep = 60 * 5;
			}
			if($progress < 95) {
				$sleep = 60 * 30;
			}
			if($progress < 80) {
				$sleep = 60 * 60;
			}
		}
		
		# check the status of the current WU
		# if it is finished, check if the F@H client, cores 
		# and mpiexec (for the SMP client) have terminated
		# if these processes have quit, reconfigure the client
		# and restart it with the new configuration
		if($current_wu{status} && $current_wu{status} eq 'finished') {
			# check every minute
			$sleep = 60;
			
			print_date("WU has finished!\n") if($cfg{verbose});

			# check if the FAH client processes have quit
			if(client_has_quit($pid)) {
			        # Configure a random userame/team combination
			        reconfigure_client();
				
			        # start the client
			        $pid = start_client();
			}
		}
		elsif($current_wu{status} && $current_wu{status} eq 'ready for upload') {
			# check every minute
			$sleep = 60;
			
			print_date("WU is ready for upload.\n") if($cfg{verbose});
		}
		else {
			print_date("Expected: $current_wu{expect_timestamp}\n") if($cfg{verbose} && $current_wu{expect_timestamp});

			print_date("WU is not finished yet.\n") if($cfg{verbose});
		}
	
		print_date("Waiting for ". ($sleep / 60) ." minutes before polling again.\n") if($cfg{verbose});

		sleep $sleep;
		
		print_log("\n") if($cfg{verbose});
	}
}

sub get_current_wu_info {
	# get the current WU information from the qd output

	my $xml = `./qd -x` || die "Error: Failed to execute './qd -x' ($!)\n";

	# parse the XML file
	my $queue_dump = XMLin( $xml,
	                        KeyAttr    => 'number',
			        ForceArray => [ 'index', 'server' ]
	                      ) || die($@);

	if($queue_dump->{error}) {
		my $error = $queue_dump->{error};
		if($error =~ s/&apos;/'/g) {}
		if($error =~ s/&gt;/>/g) {}
		if($error =~ s/&lt;/</g) {}

		print_date("Error: qd output contains an error!\n");
		print_date("$error\n");
	}

	my %current_wu = ();
	if($queue_dump->{queue} && $queue_dump->{queue}->{current_index} >= 0) {
		my $current_index = $queue_dump->{queue}->{current_index};
	
		%current_wu = %{$queue_dump->{queue}->{index}->{"$current_index"}};
	}
	
	return %current_wu;
}

sub print_usage {
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --directory <PATH>    FAH client working directory\n";
	print "-c, --client <FILE>       FAH client binary to run\n";
	print "-a, --arguments <STRING>  FAH client arguments\n";
	print "-v, --verbose             Enable verbose output\n";
	print "-h, --help                Show this usage information\n";
}

